// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "InternalSideData.h"
#include "InternalFrame.h"
#include <cstring>
#include "vpi_sidedata.h"

namespace xma {

#define XMA_SIDEDATA "SideData"

///////////////////////////////////////////////////////////////////////////////
struct SideDataMetadataMapEntry {
  const char* name;
  uint32_t*   entry;
};

//////////////////////////////////////////////////////////////////////////////
InternalSideData::InternalSideData(IntSessionFactoryPtr sf, LogPtr log, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size) : m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData(type=%d, buffer_type=%s, size=%lu)", type,
      ((buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (buffer_type == NO_BUFFER)                   ? "none" :
                                                             "unknown"),
      size);

  if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_ONLY_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData() Invalid XmaBufferType!");
  }
  m_sd = std::make_unique<XmaFrameSideData>(); // this will throw on OOM
  memset(m_sd.get(), 0, sizeof(XmaFrameSideData));

  m_sd->type                = type;
  m_sd->buffer_type         = buffer_type;
  m_sd->on_host             = (sf->GetAPIVersion() < XMA_API_VERSION_1_1) && ((buffer_type == XMA_HOST_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE));
  side_data_buffer_t buffer = m_sf->GetSideDataBuffer(type, size, buffer_type);
  if ((buffer_type == XMA_HOST_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE)) {
    m_sd->host = static_cast<uint8_t*>(vpi_get_sidedata_vaddr(buffer.get()));
  }
  if ((buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE)) {
    m_sd->device = reinterpret_cast<uint64_t>(vpi_get_sidedata_paddr(buffer.get()));
  }
  m_sd->size = size;

  bool free_t35 = (type == XMA_FRAME_SIDE_DATA_HDR10_PARAMS) || (type == XMA_FRAME_SIDE_DATA_SEI_RAW_T35_DATA);

  auto info                  = std::make_unique<PrivateSideDataInfo>(m_sf, log, std::move(buffer), this, free_t35);
  m_sd->private_do_not_touch = static_cast<void*>(info.release());

  if (m_sd->host) {
    memset(m_sd->host, 0x00, size);
    if (buffer_type == XMA_DEVICE_BUFFER_TYPE) {
      Write();
    }
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_side_data.push_back(m_sd.get());
  lock.unlock();
#endif

  IncRef(false);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalSideData::InternalSideData(IntSessionFactoryPtr sf, LogPtr log, side_data_buffer_t buffer) : m_sf(sf), m_log(log) {
  XmaBufferType buffer_type;
  if (vpi_get_sidedata_vaddr(buffer.get())) {
    if (vpi_get_sidedata_paddr(buffer.get())) {
      buffer_type = XMA_DEVICE_BUFFER_TYPE;
    } else {
      buffer_type = XMA_HOST_BUFFER_TYPE;
    }
  } else {
    if (vpi_get_sidedata_paddr(buffer.get())) {
      buffer_type = XMA_DEVICE_ONLY_BUFFER_TYPE;
    } else {
      buffer_type = NO_BUFFER;
    }
  }
  XmaFrameSideDataType type = GetXmaTypeFromVpiType(static_cast<VpiSidedataType>(vpi_get_sidedata_type(buffer.get())));
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData(type=%d, buffer_type=%s, size=%lu)", type,
      ((buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (buffer_type == NO_BUFFER)                   ? "none" :
                                                             "unknown"),
      vpi_get_sidedata_size(buffer.get()));

  if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_ONLY_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData() Invalid XmaBufferType!");
  }
  m_sd = std::make_unique<XmaFrameSideData>(); // this will throw on OOM
  memset(m_sd.get(), 0, sizeof(XmaFrameSideData));

  m_sd->type        = type;
  m_sd->buffer_type = buffer_type;
  m_sd->on_host     = (sf->GetAPIVersion() < XMA_API_VERSION_1_1) && ((buffer_type == XMA_HOST_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE));
  if ((buffer_type == XMA_HOST_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE)) {
    m_sd->host = static_cast<uint8_t*>(vpi_get_sidedata_vaddr(buffer.get()));
  }
  if ((buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE)) {
    m_sd->device = reinterpret_cast<uint64_t>(vpi_get_sidedata_paddr(buffer.get()));
  }
  m_sd->size = vpi_get_sidedata_size(buffer.get());

  auto info                  = std::make_unique<PrivateSideDataInfo>(m_sf, log, buffer, this);
  m_sd->private_do_not_touch = static_cast<void*>(info.release());

  switch (m_sd->type) {
  case XMA_FRAME_SIDE_DATA_ML_ROI:
  case XMA_FRAME_SIDE_DATA_ML_CC:
  case XMA_FRAME_SIDE_DATA_ML_CXC:
  case XMA_FRAME_SIDE_DATA_ML_LA_STAT:
  case XMA_FRAME_SIDE_DATA_ML_BOUNDING_BOX:
  case XMA_FRAME_SIDE_DATA_ML_CLASSIFICATION:
  case XMA_FRAME_SIDE_DATA_ML_YDATA:
  case XMA_FRAME_SIDE_DATA_ML_TENSOR:
  case XMA_FRAME_SIDE_DATA_DELTA_QP_MAP_H26X: {
    uint32_t                 entry_values[] = {static_cast<uint32_t>(vpi_get_sidedata_layout(buffer.get())), static_cast<uint32_t>(vpi_get_sidedata_width(buffer.get())),
        static_cast<uint32_t>(vpi_get_sidedata_height(buffer.get())), static_cast<uint32_t>(vpi_get_sidedata_batchsize(buffer.get())),
        static_cast<uint32_t>(vpi_get_sidedata_layoutch(buffer.get())), static_cast<uint32_t>(vpi_get_sidedata_dataformat(buffer.get()))};
    const uint32_t           map_size       = sizeof(entry_values) / sizeof(entry_values[0]);
    SideDataMetadataMapEntry entries[map_size] = {{XMA_SIDE_DATA_METADATA_ML_LAYOUT, &entry_values[0]}, {XMA_SIDE_DATA_METADATA_ML_WIDTH, &entry_values[1]},
        {XMA_SIDE_DATA_METADATA_ML_HEIGHT, &entry_values[2]}, {XMA_SIDE_DATA_METADATA_ML_BATCH_SIZE, &entry_values[3]},
        {XMA_SIDE_DATA_METADATA_ML_LAYOUT_CH, &entry_values[4]}, {XMA_SIDE_DATA_METADATA_ML_DATA_FORMAT, &entry_values[5]}};
    for (uint32_t map_index = 0; map_index < map_size; ++map_index) {
      XmaParameter param;
      param.name   = const_cast<char*>(entries[map_index].name);
      param.type   = XMA_UINT32;
      param.length = sizeof(uint32_t);
      param.value  = entries[map_index].entry;
      SetMetadata(&param);
    }
  } break;
  default:
    break;
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_side_data.push_back(m_sd.get());
  lock.unlock();
#endif

  IncRef(false);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalSideData::~InternalSideData() {
  if (!m_sd) {
    return;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::~InternalSideData(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  DecRef(false);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::~InternalSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
IntSideDataPtr InternalSideData::AllocSideData(IntSessionFactoryPtr sf, LogPtr log, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size) {
  IntSideDataPtr sd = std::make_shared<InternalSideData>(sf, log, type, buffer_type, size);
  if (sf->IsC()) {
    PrivateSideDataInfo* info = sd->GetPrivateInfo();
    info->SetSelf(sd);
  }
  return sd;
}
///////////////////////////////////////////////////////////////////////////////
IntSideDataPtr InternalSideData::AllocSideData(IntSessionFactoryPtr sf, LogPtr log, side_data_buffer_t buffer) {
  IntSideDataPtr sd = std::make_shared<InternalSideData>(sf, log, buffer);
  if (sf->IsC()) {
    PrivateSideDataInfo* info = sd->GetPrivateInfo();
    info->SetSelf(sd);
  }
  return sd;
}

///////////////////////////////////////////////////////////////////////////////
VpiSidedataType InternalSideData::GetVpiTypeFromXmaType(const XmaFrameSideDataType type) {
  static const VpiSidedataType vpi_types[] = {
      VPI_SB_SEI_USER_UNREGISTERED,
      VPI_SB_SEI_USER_UNREGISTERED,
      VPI_SB_VUI_PARAMS,
      VPI_SB_ML_ROI,
      VPI_SB_ML_CC,
      VPI_SB_ML_CXC,
      VPI_SB_ML_LA_STAT,
      VPI_SB_ML_BOUNDING_BOX,
      VPI_SB_ML_CLASSIFICATION,
      VPI_SB_ML_YDATA,
      VPI_SB_ML_TENSOR,
      VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME,
      VPI_SB_SEI_CONTENT_LIGHT_LEVEL,
      VPI_SB_SEI_ATC,
      VPI_SB_SEI_HDR10_PLUS,
      VPI_SB_SEI_DOLBY_VISION,
      VPI_SB_SEI_RAW_T35_DATA,
      VPI_SB_SEI_USER_UNREGISTERED,
      VPI_SB_DQPMAP_HEVC_H264,
  };
  if ((type < 0) || (static_cast<uint32_t>(type) > (sizeof(vpi_types) / sizeof(vpi_types[0])))) {
    return VPI_SB_SEI_USER_UNREGISTERED;
  }
  return vpi_types[type];
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideDataType InternalSideData::GetXmaTypeFromVpiType(const VpiSidedataType type) {
  static const XmaFrameSideDataType xma_types[] = {
      XMA_FRAME_SIDE_DATA_ML_ROI,
      XMA_FRAME_SIDE_DATA_DELTA_QP_MAP_H26X,
      XMA_FRAME_SIDE_DATA_ML_CC,
      XMA_FRAME_SIDE_DATA_ML_CXC,
      XMA_FRAME_SIDE_DATA_ML_LA_STAT,
      XMA_FRAME_SIDE_DATA_ML_BOUNDING_BOX,
      XMA_FRAME_SIDE_DATA_ML_CLASSIFICATION,
      XMA_FRAME_SIDE_DATA_ML_YDATA,
      XMA_FRAME_SIDE_DATA_ML_TENSOR,
      XMA_FRAME_SIDE_DATA_VUI_PARAMS,
      XMA_FRAME_SIDE_DATA_SEI_MASTERING_DISPLAY_COLOR_VOLUME,
      XMA_FRAME_SIDE_DATA_SEI_CONTENT_LIGHT_LEVEL,
      XMA_FRAME_SIDE_DATA_SEI_ATC,
      XMA_FRAME_SIDE_DATA_SEI_HDR10_PLUS,
      XMA_FRAME_SIDE_DATA_SEI_DOLBY_VISION,
      XMA_FRAME_SIDE_DATA_SEI_RAW_T35_DATA,
      XMA_FRAME_SIDE_DATA_SEI_USER_UNREGISTERED,
  };
  if ((type < 0) || (static_cast<uint32_t>(type) > (sizeof(xma_types) / sizeof(xma_types[0])))) {
    return XMA_FRAME_SIDE_DATA_SEI_USER_UNREGISTERED;
  }
  return xma_types[type];
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::GetRefCount() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetRefCount", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetRefCount(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetRefCount() - Returning");
  return m_sd->refcount;
}

///////////////////////////////////////////////////////////////////////////////
bool InternalSideData::IsFrozen() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::IsFrozen", XMA_SIDEDATA, m_sd);

  PrivateSideDataInfo* info = GetPrivateInfo();
  return info->IsFrozen();

  THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::IsFrozen() XmaFrameSideData was not created by xma_side_data_alloc()!");
}

///////////////////////////////////////////////////////////////////////////////
side_data_buffer_t InternalSideData::GetVpiSideData() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetVpiSideData", XMA_SIDEDATA, m_sd);

  PrivateSideDataInfo* info = GetPrivateInfo();
  return info->GetVpiSideData();

  THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetVpiSideData() XmaFrameSideData was not created by xma_side_data_alloc()!");
}

///////////////////////////////////////////////////////////////////////////////
IntSideDataPtr InternalSideData::GetIntSideData(LogPtr log, XmaFrameSideData* side_data) {
  PrivateSideDataInfo* info = GetPrivateInfo(log, side_data);
  InternalSideData*    sd   = info->GetSelfPtr();
  if (!sd) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetIntSideData() Invalid Side Data!");
  }
  return std::static_pointer_cast<InternalSideData>(sd->shared_from_this());
}

///////////////////////////////////////////////////////////////////////////////
PrivateSideDataInfo* InternalSideData::GetPrivateInfo(LogPtr log, XmaFrameSideData* side_data) {
  PrivateSideDataInfo* info = reinterpret_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
  if ((!info) || (!info->IsValid())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetPrivateInfo() Invalid Side Data!");
  }
  return info;
}

///////////////////////////////////////////////////////////////////////////////
PrivateSideDataInfo* InternalSideData::GetPrivateInfo(XmaFrameSideData* side_data) {
  PrivateSideDataInfo* info = reinterpret_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
  if ((!info) || (!info->IsValid())) {
    return nullptr;
  }
  return info;
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideData* InternalSideData::GetSideData() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetSideData", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetSideData(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetSideData() - Returning");

  return m_sd.get();
}

///////////////////////////////////////////////////////////////////////////////
PrivateSideDataInfo* InternalSideData::GetPrivateInfo() const {
  return GetPrivateInfo(m_log, m_sd.get());
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::IncRef(bool frame_holding) {
  LOCK_SIDE_DATA(m_log, "InternalSideData::IncRef", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::IncRef(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  if (frame_holding) {
    PrivateSideDataInfo* info = GetPrivateInfo();
    info->IncFramesHolding();
  }

  ++m_sd->refcount;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::IncRef() - Returning");
  return m_sd->refcount;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::DecRef(bool frame_holding) {
  LOCK_SIDE_DATA(m_log, "InternalSideData::DecRef", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::DecRef(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  if (frame_holding) {
    PrivateSideDataInfo* info = GetPrivateInfo();
    info->DecFramesHolding();
  }

  --m_sd->refcount;

  if (m_sd->refcount > 0) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::DecRef() - Returning");
    return m_sd->refcount;
  }

  PrivateSideDataInfo* info = GetPrivateInfo();

  if (m_sf->IsC()) {
    info->SetSelf(nullptr);
  }

  m_sd->host        = nullptr;
  m_sd->device      = 0;
  m_sd->buffer_type = NO_BUFFER;

  if (m_sd->metadata) {
    for (uint32_t index = 0; m_sd->metadata[index].name; ++index) {
      delete[] m_sd->metadata[index].name;
      delete[] reinterpret_cast<uint8_t*>(m_sd->metadata[index].value);
    }
    delete[] m_sd->metadata;
    m_sd->metadata = nullptr;
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  if (m_sf->m_side_data.remove(m_sd.get()) != 1) {
    m_log->LogMsg(XMA_ERROR_LOG, XMA_SIDEDATA, "XMA Memory check Error! Side Data %p removed twice?", m_sd.get());
  }
  lock.unlock();
#endif

  guard.unlock();
  delete info;
  m_sd->private_do_not_touch = nullptr;
  m_sd                       = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::DecRef() - Returning");

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
void InternalSideData::SetFrozen() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::SetFrozen", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::SetFrozen(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  PrivateSideDataInfo* info = GetPrivateInfo();
  info->SetFrozen();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::SetFrozen() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
void InternalSideData::CopyMetadata(SideDataPtr src) {
  LOCK_SIDE_DATA(m_log, "InternalSideData::CopyMetadata", XMA_SIDEDATA, m_sd);

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Unable to copy metadata when frozen!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata(type=%d)", m_sd->type);

  IntSideDataPtr int_src = std::dynamic_pointer_cast<InternalSideData>(src);
  if (!int_src) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Source XmaFrameSideData was not created by xma_side_data_alloc()!");
  }

  XmaParameter* src_metadata = int_src->m_sd->metadata;
  if (!src_metadata) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() - Returning");
    return;
  }

  uint32_t src_count  = 0;
  uint32_t dst_count  = 0;
  uint32_t duplicates = 0;
  char*    src_name;
  if (m_sd->metadata) {
    for (; (src_name = src_metadata[src_count].name); ++src_count) {
      char*    name;
      uint32_t current_count = 0;
      for (; (name = m_sd->metadata[current_count].name); ++current_count) {
        if (strcmp(src_name, name) == 0) {
          ++duplicates;
          break;
        }
      }
    }

    for (; m_sd->metadata[dst_count].name; ++dst_count)
      ;
  } else {
    for (; src_metadata[src_count].name; ++src_count)
      ;
  }

  uint32_t      new_count    = src_count + dst_count - duplicates;
  XmaParameter* dst_metadata = new XmaParameter[new_count + 1];
  if (!dst_metadata) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Out of memory!");
  }
  memset(dst_metadata, 0, sizeof(XmaParameter) * (new_count + 1));

  if (m_sd->metadata) {
    uint32_t current_count;
    for (src_count = 0; (src_name = src_metadata[src_count].name); ++src_count) {
      char* name;
      bool  found = false;
      for (current_count = 0; (name = m_sd->metadata[current_count].name); ++current_count) {
        if (strcmp(src_name, name) == 0) {
          memcpy(&dst_metadata[src_count], &m_sd->metadata[current_count], sizeof(XmaParameter));
          found = true;
          break;
        }
      }
      if (!found) {
        dst_metadata[src_count].name = new char[strlen(src_name) + 1];
        if (!dst_metadata[src_count].name) {
          THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Out of memory!");
        }
        strcpy(dst_metadata[src_count].name, src_name);
        dst_metadata[src_count].type   = src_metadata[src_count].type;
        dst_metadata[src_count].length = src_metadata[src_count].length;
        dst_metadata[src_count].value  = new unsigned char[src_metadata[src_count].length];
        if (!dst_metadata[src_count].value) {
          THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Out of memory!");
        }
        memcpy(dst_metadata[src_count].value, src_metadata[src_count].value, src_metadata[src_count].length);
      }
    }
    char* name;
    if (src_count < new_count) {
      uint32_t count = src_count;
      for (current_count = 0; (name = m_sd->metadata[current_count].name); ++current_count) {
        bool found = false;
        for (uint32_t dst_index = 0; dst_index < count; ++dst_index) {
          if (strcmp(name, dst_metadata[dst_index].name) == 0) {
            found = true;
            break;
          }
        }
        if (!found) {
          memcpy(&dst_metadata[src_count++], &m_sd->metadata[current_count], sizeof(XmaParameter));
        }
      }
    }
    delete[] m_sd->metadata;
  } else {
    for (src_count = 0; (src_name = src_metadata[src_count].name); ++src_count) {
      dst_metadata[src_count].name = new char[strlen(src_name) + 1];
      if (!dst_metadata[src_count].name) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Out of memory!");
      }
      strcpy(dst_metadata[src_count].name, src_name);
      dst_metadata[src_count].type   = src_metadata[src_count].type;
      dst_metadata[src_count].length = src_metadata[src_count].length;
      dst_metadata[src_count].value  = new unsigned char[src_metadata[src_count].length];
      if (!dst_metadata[src_count].value) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::CopyMetadata() Out of memory!");
      }
      memcpy(dst_metadata[src_count].value, src_metadata[src_count].value, src_metadata[src_count].length);
    }
  }
  m_sd->metadata = dst_metadata;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::Read() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::Read", XMA_SIDEDATA, m_sd);

  if (m_sf->GetAPIVersion() >= XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() Unsupported feature in this version of the API!");
  }

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() Unable to read when frozen!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Read(type=%d)", m_sd->type);

  if (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() XmaFrameSideData does not contain both host and device buffers!");
  }

  PrivateSideDataInfo* info   = GetPrivateInfo();
  side_data_buffer_t   buffer = info->GetVpiSideData();
  if ((vpi_sidedata_sync(buffer.get(), VPI_DATA_DWLOAD)) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() Read failed!");
  } else {
    m_sd->on_host = true;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Read() - Returning");
    return XMA_SUCCESS;
  }

  THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() XmaFrameSideData was not created by xma_side_data_alloc()!");
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideDataType InternalSideData::GetType() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetType", XMA_SIDEDATA, m_sd);

  return m_sd->type;
}

///////////////////////////////////////////////////////////////////////////////
XmaBufferType InternalSideData::GetBufferType() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetBufferType", XMA_SIDEDATA, m_sd);

  return m_sd->buffer_type;
}

///////////////////////////////////////////////////////////////////////////////
bool InternalSideData::IsOnHost() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::IsOnHost", XMA_SIDEDATA, m_sd);

  if (m_sf->GetAPIVersion() >= XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::IsOnHost() Unsupported feature in this version of the API!");
  }

  return m_sd->on_host;
}

///////////////////////////////////////////////////////////////////////////////
uint8_t* InternalSideData::GetHostPtr() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetHostPtr", XMA_SIDEDATA, m_sd);

  if ((m_sd->buffer_type != XMA_HOST_BUFFER_TYPE) && (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetHostPtr() XmaFrameSideData does not contain a host buffer!");
  }

  return m_sd->host;
}

///////////////////////////////////////////////////////////////////////////////
uint64_t InternalSideData::GetDevicePtr() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetDevicePtr", XMA_SIDEDATA, m_sd);

  if ((m_sd->buffer_type != XMA_DEVICE_ONLY_BUFFER_TYPE) && (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetDevicePtr() XmaFrameSideData does not contain a device buffer!");
  }

  return m_sd->device;
}

///////////////////////////////////////////////////////////////////////////////
size_t InternalSideData::GetSize() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetSize", XMA_SIDEDATA, m_sd);

  return m_sd->size;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::GetMetadata(XmaParameter* metadata) const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetMetadata", XMA_SIDEDATA, m_sd);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetMetadata() Unsupported feature in this version of the API!");
  }

  if (!metadata) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetMetadata() Invalid XmaParameter!");
  }

  if (!m_sd->metadata) {
    return XMA_ERROR_INVALID;
  }

  for (uint32_t index = 0; m_sd->metadata[index].name; ++index) {
    if (strcmp(metadata->name, m_sd->metadata[index].name) == 0) {
      if (metadata->type != m_sd->metadata[index].type) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetMetadata() Invalid type for metadata %s!", metadata->name);
      }
      if (metadata->length < m_sd->metadata[index].length) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetMetadata() Invalid length for metadata %s!", metadata->name);
      }
      if (metadata->value == NULL) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetMetadata() XmaParameter value must not be NULL for metadata %s!", metadata->name);
      }
      memcpy(metadata->value, m_sd->metadata[index].value, m_sd->metadata[index].length);
      metadata->length = m_sd->metadata[index].length;
      return XMA_SUCCESS;
    }
  }

  return XMA_ERROR_INVALID;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::SetMetadata(const XmaParameter* metadata) {
  LOCK_SIDE_DATA(m_log, "InternalSideData::SetMetadata", XMA_SIDEDATA, m_sd);

  if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Unsupported feature in this version of the API!");
  }

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Unable to set metadata when frozen!");
  }

  if (!metadata) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Invalid XmaParameter!");
  }

  uint32_t index = 0;
  if (m_sd->metadata) {
    for (; m_sd->metadata[index].name; ++index) {
      if (strcmp(metadata->name, m_sd->metadata[index].name) == 0) {
        if (metadata->length != m_sd->metadata[index].length) {
          delete[] reinterpret_cast<unsigned char*>(m_sd->metadata[index].value);
          m_sd->metadata[index].value = new unsigned char[metadata->length];
          if (!m_sd->metadata[index].value) {
            THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Out of memory!");
          }
        }
        m_sd->metadata[index].type   = metadata->type;
        m_sd->metadata[index].length = metadata->length;
        memcpy(m_sd->metadata[index].value, metadata->value, metadata->length);
        return XMA_SUCCESS;
      }
    }
  }

  XmaParameter* new_metadata = new XmaParameter[index + 2];
  if (!new_metadata) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Out of memory!");
  }
  memset(new_metadata, 0, sizeof(XmaParameter) * (index + 2));

  new_metadata[index].name = new char[strlen(metadata->name) + 1];
  if (!new_metadata[index].name) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Out of memory!");
  }
  strcpy(new_metadata[index].name, metadata->name);
  new_metadata[index].type   = metadata->type;
  new_metadata[index].length = metadata->length;
  new_metadata[index].value  = new unsigned char[metadata->length];
  if (!new_metadata[index].value) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::SetMetadata() Out of memory!");
  }
  memcpy(new_metadata[index].value, metadata->value, metadata->length);

  new_metadata[index + 1].name   = nullptr;
  new_metadata[index + 1].type   = static_cast<XmaDataType>(0);
  new_metadata[index + 1].length = 0;
  new_metadata[index + 1].value  = nullptr;

  if (m_sd->metadata) {
    memcpy(new_metadata, m_sd->metadata, sizeof(XmaParameter) * index);
    delete[] m_sd->metadata;
  }
  m_sd->metadata = new_metadata;

  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::Write() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::Write", XMA_SIDEDATA, m_sd);

  if (IsFrozen()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() Unable to write when frozen!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Write(type=%d)", m_sd->type);

  if (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() XmaFrameSideData does not contain both host and device buffers!");
  }

  PrivateSideDataInfo* info   = GetPrivateInfo();
  side_data_buffer_t   buffer = info->GetVpiSideData();
  if ((vpi_sidedata_sync(buffer.get(), VPI_DATA_UPLOAD)) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() Write failed!");
  } else {
    m_sd->on_host = false;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Write() - Returning");
    return XMA_SUCCESS;
  }

  THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() XmaFrameSideData was not created by xma_side_data_alloc()!");
}

///////////////////////////////////////////////////////////////////////////////
int32_t AddSideDataToVpiFrm(IntSessionFactoryPtr sf, LogPtr log, FramePtr src, VpiFrm* dst) {
  if (!src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() No available source frame!");
  }
  if (!dst) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() No available destination frame!");
  }

  log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm()");

  uint32_t sd_count[VPI_SB_TYPE_MAX];
  memset(sd_count, 0, sizeof(sd_count));

  XmaAPIVersion api_version = sf->GetAPIVersion();
  IntFramePtr   int_src     = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!int_src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid frame!");
  }
  SideDataPtr        sd;
  IntSideDataPtr     int_sd;
  side_data_buffer_t buffer;
  for (uint32_t index = 0; (sd = int_src->GetSideDataAtIndex(index)); ++index) {
    XmaBufferType buffer_type = sd->GetBufferType();
    switch (sd->GetType()) {
    case XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS:
      if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
        THROW_RUNTIME_EXCEPTION(
            log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS!");
      }
      if (api_version < XMA_API_VERSION_1_1_2) {
        if (sd->GetSize() != sizeof(XmaDynamicEncParams)) {
          THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid size for side data XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS!");
        }
        memset(&dst->dynamic_params, 0, sizeof(VpiDynamicParams));
        memcpy(&dst->dynamic_params, sd->GetHostPtr(), sizeof(XmaDynamicEncParams));
      } else {
        if (sd->GetSize() != sizeof(XmaDynamicEncParams_v2)) {
          THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid size for side data XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS!");
        }
        memcpy(&dst->dynamic_params, sd->GetHostPtr(), sizeof(XmaDynamicEncParams_v2));
      }
      break;
    case XMA_FRAME_SIDE_DATA_HDR10_PARAMS:
      if (api_version >= XMA_API_VERSION_1_1) {
        THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Unsupported feature in this version of the API!");
      } else if (sd->GetSize() != sizeof(XmaHdrParams)) {
        THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid size for side data XMA_FRAME_SIDE_DATA_HDR10_PARAMS!");
      } else {
        if (++sd_count[VPI_SB_SEI_MASTERING_DISPLAY_COLOR_VOLUME] > VPI_MAX_SB_NUM) {
          THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type XMA_FRAME_SIDE_DATA_HDR10_PARAMS!");
        }
        if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
          THROW_RUNTIME_EXCEPTION(
              log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_HDR10_PARAMS!");
        }

        XmaHdrParams* hdr_params = reinterpret_cast<XmaHdrParams*>(sd->GetHostPtr());

        int_sd = InternalSideData::AllocSideData(sf, log, XMA_FRAME_SIDE_DATA_SEI_MASTERING_DISPLAY_COLOR_VOLUME, XMA_DEVICE_BUFFER_TYPE, sizeof(XmaMasterDisColVol));
        memcpy(int_sd->GetHostPtr(), &hdr_params->mdcv, sizeof(XmaMasterDisColVol));
        int_sd->Write();
        buffer = int_sd->GetVpiSideData();
        vpifrm_attach_sidedata(dst, buffer.get());

        int_sd = InternalSideData::AllocSideData(sf, log, XMA_FRAME_SIDE_DATA_SEI_CONTENT_LIGHT_LEVEL, XMA_DEVICE_BUFFER_TYPE, sizeof(XmaContentLightLevel));
        memcpy(int_sd->GetHostPtr(), &hdr_params->cll, sizeof(XmaContentLightLevel));
        int_sd->Write();
        buffer = int_sd->GetVpiSideData();
        vpifrm_attach_sidedata(dst, buffer.get());

        int_sd = InternalSideData::AllocSideData(sf, log, XMA_FRAME_SIDE_DATA_SEI_ATC, XMA_DEVICE_BUFFER_TYPE, sizeof(XmaATCInfo));
        memcpy(int_sd->GetHostPtr(), &hdr_params->atc, sizeof(XmaATCInfo));
        int_sd->Write();
        buffer = int_sd->GetVpiSideData();
        vpifrm_attach_sidedata(dst, buffer.get());

        if (hdr_params->t35) {
          XmaT35Params* t35 = hdr_params->t35;
          for (uint8_t t35_index = 0; t35_index < t35->payload_count; ++t35_index) {
            uint8_t* data                            = t35->payload_byte_data[t35_index];
            uint16_t terminal_provided_code          = data[0] << 8 | data[1];
            uint16_t terminal_provided_oriented_code = data[2] << 8 | data[3];
            if ((terminal_provided_code == 0x3B) && (terminal_provided_oriented_code == 0x00)) {
              if (++sd_count[VPI_SB_SEI_DOLBY_VISION] > VPI_MAX_SB_NUM) {
                THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type XMA_FRAME_SIDE_DATA_SEI_DOLBY_VISION!");
              }
              int_sd = InternalSideData::AllocSideData(sf, log, XMA_FRAME_SIDE_DATA_SEI_DOLBY_VISION, XMA_DEVICE_BUFFER_TYPE, sizeof(XmaSMPTE_ST2094_10_v2));
              XmaSMPTE_ST2094_10_v2* st2094_10 = reinterpret_cast<XmaSMPTE_ST2094_10_v2*>(int_sd->GetHostPtr());
              st2094_10->country_code          = t35->country_code[t35_index];
              st2094_10->extended_country_code = t35->extended_country_code[t35_index];
              memcpy(&st2094_10->st2094_10, &t35->meta[t35_index].st2094_10, sizeof(XmaSMPTE_ST2094_10));
              int_sd->Write();
              buffer = int_sd->GetVpiSideData();
              vpifrm_attach_sidedata(dst, buffer.get());
            } else if ((terminal_provided_code == 0x3C) && (terminal_provided_oriented_code == 0x01)) {
              if (++sd_count[VPI_SB_SEI_HDR10_PLUS] > VPI_MAX_SB_NUM) {
                THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type XMA_FRAME_SIDE_DATA_SEI_HDR10_PLUS!");
              }
              int_sd = InternalSideData::AllocSideData(sf, log, XMA_FRAME_SIDE_DATA_SEI_HDR10_PLUS, XMA_DEVICE_BUFFER_TYPE, sizeof(XmaSMPTE_ST2094_40_v2));
              XmaSMPTE_ST2094_40_v2* st2094_40 = reinterpret_cast<XmaSMPTE_ST2094_40_v2*>(int_sd->GetHostPtr());
              st2094_40->country_code          = t35->country_code[t35_index];
              st2094_40->extended_country_code = t35->extended_country_code[t35_index];
              memcpy(&st2094_40->st2094_40, &t35->meta[t35_index].st2094_40, sizeof(XmaSMPTE_ST2094_40));
              int_sd->Write();
              buffer = int_sd->GetVpiSideData();
              vpifrm_attach_sidedata(dst, buffer.get());
            }
            if (++sd_count[VPI_SB_SEI_RAW_T35_DATA] > VPI_MAX_SB_NUM) {
              THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type XMA_FRAME_SIDE_DATA_SEI_RAW_T35_DATA!");
            }
            int_sd                       = InternalSideData::AllocSideData(sf, log, XMA_FRAME_SIDE_DATA_SEI_RAW_T35_DATA, XMA_DEVICE_BUFFER_TYPE, sizeof(XmaRawT35Data));
            XmaRawT35Data* raw_t35       = reinterpret_cast<XmaRawT35Data*>(int_sd->GetHostPtr());
            raw_t35->payload_byte_length = t35->payload_byte_length[t35_index];
            raw_t35->payload_byte_data   = new uint8_t[raw_t35->payload_byte_length];
            memcpy(raw_t35->payload_byte_data, t35->payload_byte_data[t35_index], raw_t35->payload_byte_length);
            int_sd->Write();
            buffer = int_sd->GetVpiSideData();
            vpifrm_attach_sidedata(dst, buffer.get());
          }
        }
      }
      break;
    case XMA_FRAME_SIDE_DATA_VUI_PARAMS:
      if (++sd_count[InternalSideData::GetVpiTypeFromXmaType(sd->GetType())] > VPI_MAX_SB_NUM) {
        THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type %u!", static_cast<uint32_t>(sd->GetType()));
      }
      int_sd = std::dynamic_pointer_cast<InternalSideData>(sd);
      buffer = int_sd->GetVpiSideData();
      vpifrm_attach_sidedata(dst, buffer.get());
      break;
    case XMA_FRAME_SIDE_DATA_ML_ROI:
    case XMA_FRAME_SIDE_DATA_ML_CC:
    case XMA_FRAME_SIDE_DATA_ML_CXC:
    case XMA_FRAME_SIDE_DATA_ML_LA_STAT:
    case XMA_FRAME_SIDE_DATA_ML_BOUNDING_BOX:
    case XMA_FRAME_SIDE_DATA_ML_CLASSIFICATION:
    case XMA_FRAME_SIDE_DATA_ML_YDATA:
    case XMA_FRAME_SIDE_DATA_ML_TENSOR:
    case XMA_FRAME_SIDE_DATA_DELTA_QP_MAP_H26X:
      if (api_version < XMA_API_VERSION_1_1) {
        THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Unsupported feature in this version of the API!");
      } else {
        if (++sd_count[InternalSideData::GetVpiTypeFromXmaType(sd->GetType())] > VPI_MAX_SB_NUM) {
          THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type %u!", static_cast<uint32_t>(sd->GetType()));
        }
        int_sd = std::dynamic_pointer_cast<InternalSideData>(sd);
        if (!int_sd) {
          THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid side data!");
        }
        buffer              = int_sd->GetVpiSideData();
        VpiSidedata* vpi_sd = reinterpret_cast<VpiSidedata*>(buffer.get());
        XmaParameter param;
        param.name   = const_cast<char*>(XMA_SIDE_DATA_METADATA_ML_LAYOUT);
        param.type   = XMA_UINT32;
        param.length = sizeof(uint32_t);
        param.value  = &vpi_sd->layout;
        int_sd->GetMetadata(&param);
        param.name  = const_cast<char*>(XMA_SIDE_DATA_METADATA_ML_DATA_FORMAT);
        param.value = &vpi_sd->data_format;
        int_sd->GetMetadata(&param);
        uint32_t* nchw[4];
        bool      set_ml = false;
        switch (vpi_sd->layout) {
        case XMA_ML_LAYOUT_NCHW:
          nchw[0] = &vpi_sd->nchw[0];
          nchw[1] = &vpi_sd->nchw[1];
          nchw[2] = &vpi_sd->nchw[2];
          nchw[3] = &vpi_sd->nchw[3];
          set_ml  = true;
          break;
        case XMA_ML_LAYOUT_NHWC:
          nchw[0] = &vpi_sd->nchw[1];
          nchw[1] = &vpi_sd->nchw[2];
          nchw[2] = &vpi_sd->nchw[0];
          nchw[3] = &vpi_sd->nchw[3];
          set_ml  = true;
          break;
        case XMA_ML_LAYOUT_CHWN:
          nchw[0] = &vpi_sd->nchw[1];
          nchw[1] = &vpi_sd->nchw[2];
          nchw[2] = &vpi_sd->nchw[3];
          nchw[3] = &vpi_sd->nchw[0];
          set_ml  = true;
          break;
        }
        if (set_ml) {
          SideDataMetadataMapEntry entries[] = {{XMA_SIDE_DATA_METADATA_ML_WIDTH, nchw[0]}, {XMA_SIDE_DATA_METADATA_ML_HEIGHT, nchw[1]},
              {XMA_SIDE_DATA_METADATA_ML_BATCH_SIZE, nchw[3]}, {XMA_SIDE_DATA_METADATA_ML_LAYOUT_CH, nchw[2]}};
          const uint32_t           map_size  = sizeof(entries) / sizeof(entries[0]);
          for (uint32_t map_index = 0; map_index < map_size; ++map_index) {
            param.name   = const_cast<char*>(entries[map_index].name);
            param.type   = XMA_UINT32;
            param.length = sizeof(uint32_t);
            param.value  = nchw[map_index];
            int_sd->GetMetadata(&param);
          }
        }
        vpifrm_attach_sidedata(dst, buffer.get());
      }
      break;
    case XMA_FRAME_SIDE_DATA_SEI_MASTERING_DISPLAY_COLOR_VOLUME:
    case XMA_FRAME_SIDE_DATA_SEI_CONTENT_LIGHT_LEVEL:
    case XMA_FRAME_SIDE_DATA_SEI_ATC:
    case XMA_FRAME_SIDE_DATA_SEI_HDR10_PLUS:
    case XMA_FRAME_SIDE_DATA_SEI_DOLBY_VISION:
    case XMA_FRAME_SIDE_DATA_SEI_RAW_T35_DATA:
    case XMA_FRAME_SIDE_DATA_SEI_USER_UNREGISTERED:
      if (api_version < XMA_API_VERSION_1_1) {
        THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Unsupported feature in this version of the API!");
      } else {
        if (++sd_count[InternalSideData::GetVpiTypeFromXmaType(sd->GetType())] > VPI_MAX_SB_NUM) {
          THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Too many side data of type %u!", static_cast<uint32_t>(sd->GetType()));
        }
        int_sd = std::dynamic_pointer_cast<InternalSideData>(sd);
        if (!int_sd) {
          THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Invalid side data!");
        }
        buffer = int_sd->GetVpiSideData();
        vpifrm_attach_sidedata(dst, buffer.get());
      }
      break;
    default:
      if (sd->GetType() < XMA_FRAME_SIDE_DATA_USER_DEFINED_START) {
        THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Unsupported feature in this version of the API!");
      }
      break;
    }
  }

  log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() - Returning");

  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
PrivateSideDataInfo::PrivateSideDataInfo(IntSessionFactoryPtr sf, LogPtr log, side_data_buffer_t buffer, InternalSideData* self, bool free_t35) :
    m_buffer(std::move(buffer)), m_self_ptr(self), m_free_t35(free_t35), m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "PrivateSideDataInfo::PrivateSideDataInfo()");
}

PrivateSideDataInfo::~PrivateSideDataInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "PrivateSideDataInfo::~PrivateSideDataInfo()");
}

bool PrivateSideDataInfo::IsValid() const {
  return (m_dummy == SIDE_DATA_DUMMY_VALUE);
}

std::recursive_mutex& PrivateSideDataInfo::GetMutex() {
  return m_mutex;
}

side_data_buffer_t PrivateSideDataInfo::GetVpiSideData() {
  return m_buffer;
}

void PrivateSideDataInfo::IncFramesHolding() {
  if (!m_sf->IsC()) {
    ++m_frames_holding;
    m_self = std::static_pointer_cast<InternalSideData>(m_self_ptr->shared_from_this());
  }
}

void PrivateSideDataInfo::DecFramesHolding() {
  if (!m_sf->IsC()) {
    if (m_frames_holding <= 0) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "PrivateSideDataInfo::DecFramesHolding() Invalid reference count!");
    }
    if (--m_frames_holding == 0) {
      m_self = nullptr;
    }
  }
}

void PrivateSideDataInfo::SetSelf(IntSideDataPtr self) {
  m_self = self;
}

InternalSideData* PrivateSideDataInfo::GetSelfPtr() const {
  return m_self_ptr;
}

bool PrivateSideDataInfo::FreeT35() const {
  return m_free_t35;
}

void PrivateSideDataInfo::SetFrozen() {
  m_frozen = true;
}

bool PrivateSideDataInfo::IsFrozen() const {
  return m_frozen;
}

IntSessionFactoryPtr PrivateSideDataInfo::GetSessionFactory() const {
  return m_sf;
}

LogPtr PrivateSideDataInfo::GetLog() const {
  return m_log;
}

}; // Namespace xma
