// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <deque>
#include <mutex>
#include <functional>
#include "xmacpp/EncoderSession.h"
#include "SessionBase.h"
#include "InternalDataBuffer.h"

namespace xma {

/* The delay will be relative to the frame rate, so a framerate of 60 fps and
 * retry delay of 10 will lead to a delay of 1/(60*10) seconds */
#define XMA_ENCODER_RETRY_DELAY 10
/* The retry count is relative to the retry delay (which is relative to the
 * framerate), so if the framerate is 60 fps and the retry count is
 * (XMA_ENCODER_RETRY_DELAY*100), the maximum delay will be 100*1/60
 * seconds */
#define XMA_ENCODER_MAX_RETRY_COUNT (XMA_ENCODER_RETRY_DELAY * 100)

//////////////////////////////////////////////////////////////////////////////
class InternalEncoderSession final : virtual public EncoderSession, virtual public SessionBase {
public:
  InternalEncoderSession(IntSessionFactoryPtr sf, XmaEncoderProperties* props);
  virtual ~InternalEncoderSession();

  virtual int32_t SetLog(LogPtr log) override;

  virtual int32_t Enqueue(queue_t& data) override;
  virtual int32_t Dequeue(queue_t& data) override;

private:
  void             InitParams(XmaEncoderProperties* props);
  void             UpdateParams(FramePtr frame);
  void             StartEncoder(FramePtr& frame);
  int32_t          Push_v1_0(FramePtr frame);
  int32_t          Pull_v1_0(IntDataBufferPtr& buf);
  int32_t          Push_v1_1(FramePtr frame);
  int32_t          Pull_v1_1(IntDataBufferPtr& buf);
  int32_t          PushFrame_v1_1(FramePtr& frame);
  IntDataBufferPtr Process_v1_1();
  int32_t          Process_ULL(IntDataBufferPtr& buf);
  int32_t          GetPacket(IntDataBufferPtr& buf);
  int32_t          PullPacket();

private:
  static void Thread(InternalEncoderSession* session);

protected:
  using lock_t           = std::unique_lock<std::mutex>;
  using vpi_mem_handle_t = std::unique_ptr<void, std::function<void(VpiMemHandle mem_ctx)>>;
  using enc_handle_t     = std::unique_ptr<void, std::function<void(void*)>>;
  using pool_handle_t    = std::unique_ptr<void, std::function<void(void*)>>;
  std::mutex                   m_mutex;
  bool                         m_use_thread = false;
  std::thread                  m_thread;
  std::condition_variable      m_thread_enqueue_cond; // thread waits for next frame
  std::condition_variable      m_thread_dequeue_cond; // thread waits to push packet
  std::condition_variable      m_wait_enqueue_cond;   // push waits to be able to push frame
  std::condition_variable      m_wait_dequeue_cond;   // pull wait for next packet
  FramePtr                     m_frame;
  vpi_mem_handle_t             m_mem_ctx;
  enc_handle_t                 m_handle;
  VpiEncInitParams             m_params;
  char                         m_param_str[4096];
  uint8_t*                     m_extra_data      = nullptr;
  int32_t*                     m_extra_data_size = nullptr;
  bool                         m_resetting       = false;
  IntDataBufferPtr             m_buf;
  std::deque<IntDataBufferPtr> m_queue;
  uint32_t                     m_input_bufferpool_size = 0;
  uint32_t                     m_max_frames_processing = 0;
  uint32_t                     m_frames_processing     = 0;
  int32_t                      m_wait                  = 0;
  bool                         m_eos_pushed            = false;
  bool                         m_end_recieved          = false;
  bool                         m_stop                  = false;
  int64_t                      m_frame_enqueued        = 0;
  int64_t                      m_frame_dequeued        = 0;
  char                         m_instance_name[64];
  ExceptionStore               m_exception;
};

using IntEncoderSessionPtr = std::shared_ptr<InternalEncoderSession>;

}; // Namespace xma
