// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <memory>
#include "xmacpp/SideData.h"
#include "InternalSessionFactory.h"
#include "vpi_memory.h"

namespace xma {
//////////////////////////////////////////////////////////////////////////////
class PrivateSideDataInfo;

//////////////////////////////////////////////////////////////////////////////
class InternalSideData final : public SideData {
public:
  InternalSideData(IntSessionFactoryPtr sf, LogPtr log, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size);
  InternalSideData(IntSessionFactoryPtr sf, LogPtr log, side_data_buffer_t buffer);
  virtual ~InternalSideData();

  static std::shared_ptr<InternalSideData> AllocSideData(IntSessionFactoryPtr sf, LogPtr log, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size);
  static std::shared_ptr<InternalSideData> AllocSideData(IntSessionFactoryPtr sf, LogPtr log, side_data_buffer_t buffer);

  static VpiSidedataType      GetVpiTypeFromXmaType(const XmaFrameSideDataType type);
  static XmaFrameSideDataType GetXmaTypeFromVpiType(const VpiSidedataType type);

  int32_t            GetRefCount() const;
  bool               IsFrozen() const;
  side_data_buffer_t GetVpiSideData() const;

  static std::shared_ptr<InternalSideData> GetIntSideData(LogPtr log, XmaFrameSideData* side_data);
  static PrivateSideDataInfo*              GetPrivateInfo(LogPtr log, XmaFrameSideData* side_data);
  static PrivateSideDataInfo*              GetPrivateInfo(XmaFrameSideData* side_data);
  XmaFrameSideData*                        GetSideData() const;
  PrivateSideDataInfo*                     GetPrivateInfo() const;

  int32_t IncRef(bool frame_holding);
  int32_t DecRef(bool frame_holding);
  void    SetFrozen();
  void    CopyMetadata(SideDataPtr src);

  virtual XmaFrameSideDataType GetType() const override;
  virtual XmaBufferType        GetBufferType() const override;
  virtual bool                 IsOnHost() const override;
  virtual uint8_t*             GetHostPtr() const override;
  virtual uint64_t             GetDevicePtr() const override;
  virtual size_t               GetSize() const override;

  virtual int32_t Read() override; // deprecated in v1.1
  virtual int32_t Write() override;

  virtual int32_t SetMetadata(const XmaParameter* metadata) override; // added in v1.1
  virtual int32_t GetMetadata(XmaParameter* metadata) const override; // added in v1.1

protected:
  std::unique_ptr<XmaFrameSideData> m_sd;
  IntSessionFactoryPtr              m_sf;
  LogPtr                            m_log;
};

using IntSideDataPtr = std::shared_ptr<InternalSideData>;

int32_t AddSideDataToVpiFrm(IntSessionFactoryPtr sf, LogPtr log, FramePtr src, VpiFrm* dst);

//////////////////////////////////////////////////////////////////////////////
#define SIDE_DATA_DUMMY_VALUE 0xCDCDCDCD
class PrivateSideDataInfo {
public:
  PrivateSideDataInfo(IntSessionFactoryPtr sf, LogPtr log, side_data_buffer_t buffer, InternalSideData* self, bool free_t35 = false);
  ~PrivateSideDataInfo();

  bool                  IsValid() const;
  std::recursive_mutex& GetMutex();
  side_data_buffer_t    GetVpiSideData();
  void                  IncFramesHolding();
  void                  DecFramesHolding();
  void                  SetSelf(IntSideDataPtr self);
  InternalSideData*     GetSelfPtr() const;
  bool                  FreeT35() const;
  void                  SetFrozen();
  bool                  IsFrozen() const;
  IntSessionFactoryPtr  GetSessionFactory() const;
  LogPtr                GetLog() const;

private:
  uint32_t                     m_dummy = SIDE_DATA_DUMMY_VALUE;
  mutable std::recursive_mutex m_mutex;
  side_data_buffer_t           m_buffer;
  uint32_t                     m_frames_holding = 0;
  IntSideDataPtr               m_self;
  InternalSideData*            m_self_ptr;
  bool                         m_free_t35;
  IntSessionFactoryPtr         m_sf;
  bool                         m_frozen = false;
  LogPtr                       m_log;
};

#define LOCK_SIDE_DATA(log, func, module, side_data) \
  if (!side_data) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() XmaFrameSideData is not valid!", func); \
  } \
  xma::PrivateSideDataInfo* private_side_data_info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch); \
  if ((!private_side_data_info) || (!private_side_data_info->IsValid())) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() Private info is not valid!", func); \
  } \
  std::unique_lock<std::recursive_mutex> guard(private_side_data_info->GetMutex());

}; // Namespace xma
